@@ -48,20 +48,31 @@
 #include <exevents.h>		/* Needed for InitValuator/Proximity stuff */
 #include <X11/keysym.h>
 #include <mipointer.h>
+#include <randrstr.h>
 
 #include <tslib.h>
+#include <sys/time.h>
+#include <time.h>
 
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+#include <xserver-properties.h>
+#endif
 #ifdef XFree86LOADER
 #include <xf86Module.h>
 #endif
 
-#define MAXBUTTONS 1
+#define MAXBUTTONS 3
+#define TIME23RDBUTTON 0.5
+#define MOVEMENT23RDBUTTON 4
 
 #define DEFAULT_HEIGHT		240
 #define DEFAULT_WIDTH		320
 
 enum { TSLIB_ROTATE_NONE=0, TSLIB_ROTATE_CW=270, TSLIB_ROTATE_UD=180, TSLIB_ROTATE_CCW=90 };
 
+enum button_state { BUTTON_NOT_PRESSED = 0, BUTTON_1_PRESSED = 1, BUTTON_3_CLICK = 3, BUTTON_3_CLICKED=4, BUTTON_EMULATION_OFF=-1 };
+
 struct ts_priv {
 	XISBuffer *buffer;
 	struct tsdev *ts;
@@ -70,11 +81,9 @@ struct ts_priv {
 	int rotate;
 	int height;
 	int width;
-};
-
-static const char *DEFAULTS[] = {
-	"TslibDevice", "/dev/event0",
-	NULL
+	enum button_state state;
+	struct timeval button_down_start;
+	int button_down_x,button_down_y;
 };
 
 static void
@@ -116,63 +125,192 @@ ConvertProc( LocalDevicePtr local,
 	return TRUE;
 }
 
+struct timeval TimevalDiff(struct timeval a, struct timeval b)
+{
+	struct timeval t;
+	t.tv_sec = a.tv_sec-b.tv_sec;
+	t.tv_usec = a.tv_usec - b.tv_usec;
+	if (t.tv_usec < 0) {
+		t.tv_sec--;
+		t.tv_usec += 1000000;
+	}
+    return t;
+}
+
 static void ReadInput (LocalDevicePtr local)
 {
 	struct ts_priv *priv = (struct ts_priv *) (local->private);
 	struct ts_sample samp;
 	int ret;
 	int x,y;
+	ScrnInfoPtr pScrn = xf86Screens[priv->screen_num];
+	Rotation rotation = rrGetScrPriv (pScrn->pScreen) ? RRGetRotation(pScrn->pScreen) : RR_Rotate_0;
+	struct timeval now;
 
-	ret = ts_read(priv->ts, &samp, 1);
+	while((ret = ts_read(priv->ts, &samp, 1)) == 1) {
+		gettimeofday(&now, NULL);
+		struct timeval pressureTime = TimevalDiff(now,priv->button_down_start);
+
+		if(samp.pressure) {
+			int tmp_x = samp.x;
+
+			switch(priv->rotate) {
+			case TSLIB_ROTATE_CW:	samp.x = samp.y;
+						samp.y = priv->width - tmp_x;
+						break;
+			case TSLIB_ROTATE_UD:	samp.x = priv->width - samp.x;
+						samp.y = priv->height - samp.y;
+						break;
+			case TSLIB_ROTATE_CCW:	samp.x = priv->height - samp.y;
+						samp.y = tmp_x;
+						break;
+			default:		break;
+			}
+
+			tmp_x = samp.x;
+
+			switch(rotation) {
+			case RR_Rotate_90:
+				samp.x = (priv->height - samp.y - 1) * priv->width / priv->height;
+				samp.y = tmp_x * priv->height / priv->width;
+				break;
+			case RR_Rotate_180:
+				samp.x = priv->width - samp.x - 1;
+				samp.y = priv->height - samp.y - 1;
+				break;
+			case RR_Rotate_270:
+				samp.x = samp.y * priv->width / priv->height;
+				samp.y = (priv->width - tmp_x - 1) * priv->height / priv->width;
+				break;
+			}
+
+			priv->lastx = samp.x;
+			priv->lasty = samp.y;
+			x = samp.x;
+			y = samp.y;
+
+			xf86XInputSetScreen(local, priv->screen_num,
+					samp.x,
+					samp.y);
+
+			xf86PostMotionEvent (local->dev, TRUE, 0, 2,
+					x, y);
+
+		}
+
+		/* button pressed state machine
+		 * if pressed than press button 1, start timer and remember the tab position
+		 * if pressed longer than TIME23RDBUTTON and it is not moved more than MOVEMENT23RDBUTTON release button 1 and click button 3
+		 * if still pressed do nothing until the pressure is released
+		 */
+		switch (priv->state) {
+			 case BUTTON_EMULATION_OFF :
+				 if(priv->lastp != samp.pressure) {
+					 priv->lastp = samp.pressure;
+					 xf86PostButtonEvent(local->dev, TRUE,
+						 1, !!samp.pressure, 0, 2,
+						 priv->lastx,
+						 priv->lasty);
+				 }
+				 break;
+			case BUTTON_NOT_PRESSED :
+				if (samp.pressure) {
+					priv->button_down_start = now;
+					priv->button_down_y = samp.y;
+					priv->button_down_x = samp.x;
+					priv->state = BUTTON_1_PRESSED;
+					//ErrorF("b1 down");
+					xf86PostButtonEvent(local->dev, TRUE,
+						priv->state, TRUE, 0, 2,
+						priv->lastx,
+						priv->lasty);
+				}
+				break;
+			case BUTTON_1_PRESSED :
+				if (samp.pressure) {
+					//ErrorF("%d %d ",pressureTime.tv_sec,pressureTime.tv_usec);
+					if ((((double)pressureTime.tv_sec)+(((double)pressureTime.tv_usec)*1e-6) > TIME23RDBUTTON) &&
+					   (abs(priv->lastx-priv->button_down_x) < MOVEMENT23RDBUTTON &&
+					    abs(priv->lasty-priv->button_down_y) < MOVEMENT23RDBUTTON))
+					{
+						//ErrorF("b1 up");
+						xf86PostButtonEvent(local->dev, TRUE,
+							priv->state, FALSE, 0, 2,
+							priv->lastx,
+							priv->lasty);
+						priv->state = BUTTON_3_CLICK;
+						//ErrorF("b3 down");
+						xf86PostButtonEvent(local->dev, TRUE,
+							priv->state, TRUE, 0, 2,
+							priv->lastx,
+							priv->lasty);
+					}
+					if (abs(priv->lastx-priv->button_down_x) > MOVEMENT23RDBUTTON ||
+					    abs(priv->lasty-priv->button_down_y) > MOVEMENT23RDBUTTON) {
+						priv->button_down_start = now;
+						priv->button_down_y = samp.y;
+						priv->button_down_x = samp.x;
+						//ErrorF("b1 state reset");
+					}
+				} else {
+					//ErrorF("b1 up");
+					xf86PostButtonEvent(local->dev, TRUE,
+						priv->state, FALSE, 0, 2,
+						priv->lastx,
+						priv->lasty);
+					priv->state = BUTTON_NOT_PRESSED;
+				}
+				break;
+			case BUTTON_3_CLICK :
+				//ErrorF("b3 up");
+				xf86PostButtonEvent(local->dev, TRUE,
+					priv->state, FALSE, 0, 2,
+					priv->lastx,
+					priv->lasty);
+				priv->state = BUTTON_3_CLICKED;
+				break;
+			case BUTTON_3_CLICKED :
+				if (!samp.pressure) {
+					//ErrorF("b3 free");
+					priv->state = BUTTON_NOT_PRESSED;
+				}
+				break;
+		}
+	}
 
 	if (ret < 0) {
 		ErrorF("ts_read failed\n");
 		return;
 	}
 
-//	ErrorF("%ld.%06ld: %6d %6d %6d\n", samp.tv.tv_sec, samp.tv.tv_usec, samp.x, samp.y, samp.pressure);
-
-	if(samp.pressure) {
-		int tmp_x = samp.x;
-
-		switch(priv->rotate) {
-		case TSLIB_ROTATE_CW:	samp.x = samp.y;
-					samp.y = priv->width - tmp_x;
-					break;
-		case TSLIB_ROTATE_UD:	samp.x = priv->width - samp.x;
-					samp.y = priv->height - samp.y;
-					break;
-		case TSLIB_ROTATE_CCW:	samp.x = priv->height - samp.y;
-					samp.y = tmp_x;
-					break;
-		default:		break;
-		}
-
-		priv->lastx = samp.x;
-		priv->lasty = samp.y;
-		x = samp.x;
-		y = samp.y;
-
-		xf86XInputSetScreen(local, priv->screen_num,
-				samp.x,
-				samp.y);
-
-		xf86PostMotionEvent (local->dev, TRUE, 0, 2,
-				x, y);
-
-	}
-
-	if(priv->lastp != samp.pressure) {
-		priv->lastp = samp.pressure;
-
-		xf86PostButtonEvent(local->dev, TRUE,
-			1, !!samp.pressure, 0, 2,
-			priv->lastx,
-			priv->lasty);
-	}
-
 }
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+static void xf86TslibInitButtonLabels(Atom *labels, int nlabels)
+{
+	memset(labels, 0, nlabels * sizeof(Atom));
+	switch(nlabels)
+	{
+		default:
+		case 7:
+			labels[6] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_RIGHT);
+		case 6:
+			labels[5] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_LEFT);
+		case 5:
+			labels[4] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_DOWN);
+		case 4:
+			labels[3] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_UP);
+		case 3:
+			labels[2] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_RIGHT);
+		case 2:
+			labels[1] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_MIDDLE);
+		case 1:
+			labels[0] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_LEFT);
+			break;
+	}
+}
+#endif
+
 /*
  * xf86TslibControlProc --
  *
@@ -183,7 +321,10 @@ xf86TslibControlProc(DeviceIntPtr device, int what)
 {
 	InputInfoPtr pInfo;
 	unsigned char map[MAXBUTTONS + 1];
-	int i;
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+	Atom labels[MAXBUTTONS];
+#endif
+	int i, axiswidth, axisheight;
 	struct ts_priv *priv;
 
 	ErrorF("%s\n", __FUNCTION__);
@@ -197,32 +338,63 @@ xf86TslibControlProc(DeviceIntPtr device, int what)
 		for (i = 0; i < MAXBUTTONS; i++) {
 			map[i + 1] = i + 1;
 		}
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+		xf86TslibInitButtonLabels(labels, MAXBUTTONS);
+#endif
 
-		if (InitButtonClassDeviceStruct(device,
-						MAXBUTTONS, map) == FALSE) {
+		if (InitButtonClassDeviceStruct(device, MAXBUTTONS,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+						labels,
+#endif
+						map) == FALSE) {
 			ErrorF("unable to allocate Button class device\n");
 			return !Success;
 		}
 
 		if (InitValuatorClassDeviceStruct(device,
 						  2,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+						labels,
+#endif
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
 						  xf86GetMotionEvents,
+#endif
 						  0, Absolute) == FALSE) {
 			ErrorF("unable to allocate Valuator class device\n");
 			return !Success;
 		}
 
-		InitValuatorAxisStruct(device, 0, 0,    	/* min val */
-					       priv->width - 1,	/* max val */
-					       priv->width,	/* resolution */
-					       0,		/* min_res */
-					       priv->width);	/* max_res */
+		switch(priv->rotate) {
+		case TSLIB_ROTATE_CW:
+		case TSLIB_ROTATE_CCW:
+			axiswidth = priv->height;
+			axisheight = priv->width;
+			break;
+		default:
+			axiswidth = priv->width;
+			axisheight = priv->height;
+			break;
+		}
 
-		InitValuatorAxisStruct(device, 1, 0,    	/* min val */
-					       priv->height - 1,/* max val */
-					       priv->height,	/* resolution */
+		InitValuatorAxisStruct(device, 0,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+					       XIGetKnownProperty(AXIS_LABEL_PROP_ABS_X),
+#endif
+					       0,		/* min val */
+					       axiswidth - 1,	/* max val */
+					       axiswidth,	/* resolution */
 					       0,		/* min_res */
-					       priv->height);	/* max_res */
+					       axiswidth);	/* max_res */
+
+		InitValuatorAxisStruct(device, 1,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+					       XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Y),
+#endif
+					       0,		/* min val */
+					       axisheight - 1,	/* max val */
+					       axisheight,	/* resolution */
+					       0,		/* min_res */
+					       axisheight);	/* max_res */
 
 		if (InitProximityClassDeviceStruct (device) == FALSE) {
 			ErrorF ("Unable to allocate EVTouch touchscreen ProximityClassDeviceStruct\n");
@@ -258,9 +430,12 @@ xf86TslibControlProc(DeviceIntPtr device, int what)
 static void
 xf86TslibUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 {
+	struct ts_priv *priv = (struct ts_priv *)(pInfo->private);
 	ErrorF("%s\n", __FUNCTION__);
 	xf86TslibControlProc(pInfo->dev, DEVICE_OFF);
+	ts_close(priv->ts);
 	xfree(pInfo->private);
+	xf86DeleteInput(pInfo, 0);
 }
 
 /*
@@ -308,16 +483,16 @@ xf86TslibInit(InputDriverPtr drv, IDevPtr dev, int flags)
 	pInfo->private = priv;
 
 	/* Collect the options, and process the common options. */
-	xf86CollectInputOptions(pInfo, DEFAULTS, NULL);
+	xf86CollectInputOptions(pInfo, NULL, NULL);
 	xf86ProcessCommonOptions(pInfo, pInfo->options);
 
 	priv->screen_num = xf86SetIntOption(pInfo->options, "ScreenNumber", 0 );
 
 	priv->width = xf86SetIntOption(pInfo->options, "Width", 0);
-	if (priv->width <= 0)	priv->width = DEFAULT_WIDTH;
+	if (priv->width <= 0)	priv->width = screenInfo.screens[0]->width;
 
 	priv->height = xf86SetIntOption(pInfo->options, "Height", 0);
-	if (priv->height <= 0)	priv->height = DEFAULT_HEIGHT;
+	if (priv->height <= 0)	priv->height = screenInfo.screens[0]->height;
 
 	s = xf86SetStrOption(pInfo->options, "Rotate", 0);
 	if (s > 0) {
@@ -334,23 +509,32 @@ xf86TslibInit(InputDriverPtr drv, IDevPtr dev, int flags)
 		priv->rotate = TSLIB_ROTATE_NONE;
 	}
 
-	s = xf86SetStrOption(pInfo->options, "TslibDevice", NULL);
+ 	s = xf86CheckStrOption(dev->commonOptions, "path", NULL);
+  	if (!s)
+		s = xf86CheckStrOption(dev->commonOptions, "Device", NULL);
+ 
+	priv->ts = ts_open(s, 1);
+	xfree(s);
 
-	priv->ts = ts_open(s, 0);
 	if (!priv->ts) {
 		ErrorF("ts_open failed (device=%s)\n",s);
+		xf86DeleteInput(pInfo, 0);
 		return NULL;
 	}
 
-	xfree(s);
-
 	if (ts_config(priv->ts)) {
 		ErrorF("ts_config failed\n");
+		xf86DeleteInput(pInfo, 0);
 		return NULL;
 	}
 
 	pInfo->fd = ts_fd(priv->ts);
 
+	priv->state = BUTTON_NOT_PRESSED;
+	if (xf86SetIntOption(pInfo->options, "EmulateRightButton", 0) == 0) {
+		priv->state = BUTTON_EMULATION_OFF;
+	}
+
 	/* Mark the device configured */
 	pInfo->flags |= XI86_CONFIGURED;
 
